"""
Django Deprecated Field - A field that marks model fields as deprecated.
"""

__version__ = "0.1.1"

from .deprecated_field import DeprecatedField, deprecated

__all__ = ["DeprecatedField", "deprecated"]
